﻿:Class RunDemo
⍝ Class for doing demonstrations.

⍝ To use:
⍝ ⎕SE.SALT.Load 'lib\RunDemo -target=⎕SE'
⍝ The use is e.g. ⎕SE.RunDemo.Load ''

⍝ A demo file consists in a series of statements to be
⍝ executed initially in sequence up to a line starting with &!.
⍝ For example, the file \demo.txt could contain the lines:

⍝⍝ Demo for demo: Prepare the form
⍝ 'f' #.⎕wc 'form' 'Demo demo' ('size' (21 21))
⍝ ⎕←'Close the form'
⍝&! ⍝ script executed initially up to here
⍝ ⎕ex 'f' ⍝ this line executed when 'next' is used

⍝ By doing ⎕se.RunDemo.Load '\demo.txt' the lines will appear
⍝ in sequence and can be executed.

    :Field Public Shared Pat ⍝ default dir
    :Field Public Shared Script←''
    :Field Public Shared Posn←0
    :Field Public Shared File ⍝ The opened file

    ∇ Edit;script
      :Access Public Shared
      script←Script
      ⎕ED'script'
      Script←File.NestedText←script
    ∇

    ∇ Init;s_ ⍝ Initialize the demo
      :Access Public Shared
     ⍝ execute each line in the file up to '&!'
      :If (⍴Script)≥Posn←(2↑¨Script)⍳⊂'&!'
          :For s_ :In (Posn-1)↑Script
              ⍎s_ ⍝ ⎕IO 1 - if this is changed problems will arise
          :EndFor
      :Else ⋄ Posn←1
      :EndIf
      Next
      'Script Initialized...'
    ∇

    ∇ UnLoad ⍝ Remove the button from the menu bar
      :Access Public Shared
      {⎕EX'⎕SE.mb.',⍵}¨'Next' 'Prev' 'Init' 'Edit'
      ⎕SE.⎕EX'Demo'
    ∇

    ∇ Load name;z;file
      :Access Public Shared
     ⍝ Prepare the script to demo.
     ⍝ 'name' can be the name of a file (a string), be empty (the name will be asked) or already be a
     ⍝ series of statements to execute (a VTV).
      :If 0∊⍴Path←2 ⎕NQ'.' 'GetEnvironment' 'HOME ' ⍝ Unix
          Path←2 ⎕NQ'.' 'GetEnvironment' 'homepath' ⍝ Windows
      :EndIf
      Path,←'\Documents\'
     
      :If 0=⎕NC'⎕SE.mb.Next' ⍝ load the button on the menu bar if necessary
          '⎕SE.mb.Next'⎕WC'menuitem' 'Next'('Event' 'Select' 'Next')('Accelerator'(123 0)) ⍝ F12
          '⎕SE.mb.Prev'⎕WC'menuitem' 'Prev'('Event' 'Select' 'Prev')('Accelerator'(122 0)) ⍝ F11
          '⎕SE.mb.Init'⎕WC'menuitem' 'Init'('Event' 'Select' 'Init')
          '⎕SE.mb.Edit'⎕WC'menuitem' 'Edit'('Event' 'Select' 'Edit')
      :EndIf
     
      :Repeat
          :If 326∊⎕DR name
              Script←name ⋄ File←⎕NS'' ⍝ done
          :Else ⍝ filename: get the script from the file
              :If 0≠⍴name
                  file←(('\:'∧.≠2↑name)/Path),name,('.txt'≢¯4↑name)/'.txt'
              :Else ⍝ name not supplied: ask it
                  'F'⎕WC'FileBox' 'Load Demo Script'Path('Event' 'FileBoxOK' 1)
                  :If 3>⍴z←⎕DQ'F'
                  :OrIf 'FileBoxOK'≢2⊃z
                      →0
                  :EndIf
                  file←3⊃z
              :EndIf
              z←⎕NEW ⎕SE.UnicodeFile file
              Script←z.NestedText
              File←z
          :EndIf
          name←''
      :Until 0≠⍴z.Raw
     
      ⎕←'Loaded Script ',file
      Init
      ⎕NQ'⎕SE' 'GotFocus'
    ∇

    ∇ Next;cl;line;keys;lab;snip
      :Access Public Shared
      'cl'⎕WC'clipboard'
     Line:
      line←Posn⊃Script,⊂'⍝ That''s all folks!'
      :If '&'≡⊃line
          :Select 2⊃line,'?'
          :Case '←'          ⍝ press key(s)
              ⍎'{2 ⎕nq ''⎕se'' 22 ⍵}',2↓line
              Posn+←1
              →0⍝Line
          :Case '→'          ⍝ branch to script label
              lab←2↓line~' '
              Posn←(Script~¨' ')⍳⊂'&:',lab
              →Line
          :CaseList ':⍝'
              Posn+←1
              →Line
          :Case '+'          ⍝ overwrite red line
              line←2↓line
              2 ⎕NQ'⎕SE' 22 'QT' ⍝ Clear 1
          :Case '-'          ⍝ invisibly execute
              ⍎2↓line
              Posn+←1
              →Line
          :Case '='          ⍝ insert snippet
              Posn+←1
              snip←⍬
              :While '&='≢2↑line←Posn⊃Script
                  snip,←⊂line
                  Posn+←1
              :EndWhile
              cl.Text←snip
              →paste
          :Case '!'
              line←2↓line
          :EndSelect
      :EndIf ⍝ '&'≡⊃line
     
      :If 0=⍴line~' '
          Posn+←1
          →Line
      :EndIf
     
      'cl'⎕WS'Text'line
     
     paste:
      2 ⎕NQ'⎕se' 22 'PT'
      Posn←(1+⍴Script)⌊Posn+1
      '⎕SE.cbbot.bandsb2.sb.ml'⎕WS'Text'(Posn⊃Script,⊂'[No More]')
      ⎕SE.cbbot.bandsb2.sb.ml.Caption←''
    ∇

    ∇ Prev;cl
      :Access Public Shared
      'cl'⎕WC'clipboard'
      Posn←1⌈Posn-2
      'cl'⎕WS'Text'(Posn⊃Script)
      2 ⎕NQ'⎕se' 22 'PT'
      Posn+←1
      '⎕SE.cbbot.bandsb2.sb.ml'⎕WS'Text'(Posn⊃Script,⊂'[No More]')
      ⎕SE.cbbot.bandsb2.sb.ml.Caption←''
    ∇

:EndClass ⍝ Demo