﻿:Class IniFiles

⍝ <h2>APL-like INI files
⍝ <h3>Overview
⍝ This class provides a kind of APL-like INI files. 
⍝ <pre>
⍝ MyInstance1←⎕NEW IniFiles (⊂,'MyIniFile.ini')
⍝ MyInstance2←⎕NEW IniFiles ('MyIniFile1.ini' 'MyIniFile2.ini')
⍝ AsNamespace←MyInstance2.Convert ⎕NS ''
⍝ </pre>
⍝ For details refer to the home page
⍝ Author: Kai Jaeger ⋄ APL Team Ltd ⋄ http://aplteam.com
⍝ Homepage: http://aplwiki.com/IniFiles

    ⎕io←1 ⋄ ⎕ml←2
    :Include APLTreeUtils  ⍝∇:require =/apltreeutils

    ∇ r←Version
      :Access Public Shared
      r←(⍕⎕THIS)'2.3.0' '2012-09-14'
      ⍝ Documentation improved
      ⍝ Checks added for invalid section and/or value names (no blanks allowed!)
    ∇

    ∇ r←History
      :Access Public Shared
      r←'See http://aplwiki.com/IniFiles/ProjectPage'
    ∇

    ∇ (r data allRemarks)←Import(data allRemarks);il;refToIniFile;fullName;temp;buffer;bool;ilRemarks;i;thisRemark;this
      bool←∨/¨'!Import'∘⍷¨data                  ⍝ Where are the files to be imported?
      il←bool⌿data                              ⍝ Create it (Import List)
      ilRemarks←bool/allRemarks                 ⍝ Remarks belonging to !Import
      '!Import statement must not have a remark'⎕SIGNAL 911/⍨0∨.<↑∘⍴¨ilRemarks
      data←(~bool)/data                         ⍝ Remove the "!Import" lines from "data"
      allRemarks←(~bool)/allRemarks             ⍝ Remove the "!Import" lines from "allRemarks"
      il←(+/∧\' '=⊃il)↓¨il                      ⍝ Drop any leading blanks
      il←Lowercase¨(il⍳¨' ')↓¨il                ⍝ Drop method name, leaving the filenames
      refToIniFile←↑1⊃⎕CLASS ⎕THIS              ⍝ Ref to itself
      r←0 3⍴⊂''
      :For this :In il
          :If (~':'∊this)∧~(1⍴this)∊'/\'        ⍝ Does not start with / and has no drive letter?
              fullName←(↑⎕CMD'cd'),'\',this     ⍝ Then make it absolutle
          :Else
              fullName←this                     ⍝ Is already absolute
          :EndIf
          temp←⎕NEW refToIniFile(,⊂fullName)    ⍝ Create an instance
          r⍪←{⍵[;2 3 3]⌿⍨0=↑∘⍴¨⍵[;1]}(temp.Get ⍬ ⍬)  ⍝ Remove section data
      :EndFor
      _import←0<1⊃⍴r
    ∇

    :Property IniFilename
    :Access Public
        ∇ r←get
          r←{0∊⍴⍵:1⊃⍵ ⋄ ⊃{⍺,';',⍵}/⍵}_IniFilename
        ∇
    :EndProperty

    :Property Default
    :Access Public
    ⍝ This property is <b>initialised</b>! If set, indexing an unknown key will _
    ⍝ return the value of this property.
        ∇ r←get
          'No default defined!'⎕SIGNAL 6/⍨0=⎕NC'_default'
          r←_default
        ∇
        ∇ set arg
          _default←arg.NewValue
        ∇
    :EndProperty

    :Property OldStyleFlag
    ⍝ Read-only property that tells you whether the current INI file is old-fashioned or not.
    :Access Public
        ∇ r←get
          r←_oldStyleFlag
        ∇
    :EndProperty

    :Property Changed
    ⍝ Read-only property that tells you that the data from the INI file was changed later on by assignment.
    ⍝ Note that those changes will persist only if the ""Save"" method is invoked at some point.
    :Access Public
        ∇ r←get
          r←_changed
        ∇
    :EndProperty

    :Property keyed default data
    ⍝ This property allows external access to the INI data via indexing.
        ∇ set val;where;EndHold;keys;section;key;data;where2;i;buffer;buffer2;ref;name;line;KEY;TOC;toc;this;new;newItem;comments
          keys←{2=≡⍵:⍵ ⋄ ↑⍵}val.Indexers
          'Length Error'⎕SIGNAL(≠/↑∘⍴¨,¨keys(,val.NewValue))/11
          keys←{~':'∊⍵:⍵ ⋄ ⍵{((⍵-1)↑⍺)(⍵↓⍺)}⍵⍳':'}¨keys
          :For i :In ⍳⍴,keys
              buffer←i⊃,keys
              'Use "Section:key" syntax'⎕SIGNAL 11/⍨2≠⍴buffer
              (section key)←buffer
              section←'_',section
              :If (⍴_Sections)<where←(Uppercase _Sections)⍳⊂Uppercase section
                       ⍝ Event the section is new!
                  (Uppercase section)⎕NS''
                  _Sections,←⊂section
                  {⍵._Data←0 4⍴⍬ ⋄ ⍵.generalRemarks←''}⍎Uppercase section
                  _changed←1
                  where←⍴_Sections
              :EndIf
              data←{⍵._Data}ref←⍎Uppercase where⊃_Sections
              KEY←Uppercase key
              TOC←Uppercase toc←data[;1]
              :If (1⊃⍴data)≥where2←TOC⍳⊂KEY
              ⍝ It is an update!
                  buffer←i⊃,val.NewValue
                  'Invalid depth'⎕SIGNAL 11/⍨~0 1 2∊⍨≡buffer
                  :If 2=≡buffer
                      new←''
                      :For this :In buffer
                          :If IsChar this
                              newItem←⊂ReplaceCurlies toc this
                              newItem,←⊂'''',(DoubleQuotes this),''''
                              new,←⊂newItem
                          :Else
                              new,←⊂2⍴⊂⍕this
                          :EndIf
                      :EndFor
                      _changed∨←data[where2;2]≢1⊃¨new
                      :If (1+⍴new)≠⍴4⊃data[where2;] ⍝ Add one for initialiation (nested!)
                      :AndIf ~0∊⍴4⊃data[where2;]
                          ⍝ The length has changed, so the comments do not any longer fit!
                          comments←(1,(2⊃data[where2;])∊(1⊃¨new))/4⊃data[where2;]
                          comments←(1,(1⊃¨new)∊2⊃data[where2;])\comments
                          data[where2;4]←⊂comments
                      :EndIf
                      data[where2;2 3]←↓⍉⊃new
                  :Else
                      :If '{'∊buffer
                          data[where2;2]←⊂ReplaceCurlies toc buffer
                          data[where2;3]←⊂'''',(DoubleQuotes buffer),''''
                      :EndIf
                      _changed∨←data[where2;2]≢⊂buffer
                      data[where2;2]←⊂DoubleQuotes buffer
                  :EndIf
              :Else
              ⍝ key is new in this section
                  buffer←i⊃,val.NewValue
                  'Invalid depth'⎕SIGNAL 11/⍨~0 1 2∊⍨≡buffer
                  :If 2=≡buffer
                      new←''
                      :For this :In buffer
                          :If IsChar this
                              newItem←⊂ReplaceCurlies toc this
                              newItem,←⊂'''',(DoubleQuotes this),''''
                          :Else
                              newItem←2⍴⊂⍕this
                          :EndIf
                          new,←⊂newItem
                      :EndFor
                      _changed←1
                      data⍪←(⊂key),(↓⍉⊃new),⊂''
                  :Else
                      :If 2=+/'{}'∊buffer
                          name←{{⍵↑⍨¯1+'}'⍳⍨⍵}⍵↓⍨⍵⍳'{'}buffer
                          :If (1⊃⍴ref._Data)<line←(Uppercase ref._Data[;1])⍳⊂Uppercase name
                              6 ⎕SIGNAL⍨'Unknown: "',name,'"'
                          :EndIf
                          buffer2←(line⊃ref._Data[;2]){a←¯1+⍵⍳'{' ⋄ w←a⌽⍵ ⋄ (-a)⌽⍺,⍵↓⍨⍵⍳'}'}buffer
                      :Else
                          buffer2←buffer
                      :EndIf
                      :If IsChar buffer
                          buffer←DoubleQuotes buffer
                          buffer←'''',(⍕buffer),''''
                      :Else
                          buffer←{0 1∊⍨≡⍵:⍕⍵ ⋄ ∇¨⍵}buffer
                      :EndIf
                      data←data,[1]key buffer2 buffer''  ⍝ Avoid ,← syntax here - bug in Dyalog 2009-07
                      _changed←1
                  :EndIf
              :EndIf
              data{⍵._Data←⍺}⍎Uppercase where⊃_Sections
          :EndFor
        ∇

        ∇ r←get arg;keys;bool;where;this;KEYS;THIS;i;noOf
          'Invalid argument'⎕SIGNAL 11/⍨0 1∊⍨≡keys←↑arg.Indexers
          :If 0∊⍴keys←{~':'∊⍵:⍵ ⋄ ⍵{((⍵-1)↑⍺)(⍵↓⍺)}⍵⍳':'}¨keys
              r←''
          :Else
              'Invalid key'⎕SIGNAL 11/⍨0=⍴∊keys
              KEYS←Uppercase¨keys
              :If 0∊bool←('_',¨1⊃¨KEYS)∊Uppercase _Sections
                  11 ⎕SIGNAL⍨'Unknown section: ',⊃{⍺,',',⍵}/∪(~bool)/1⊃¨keys
              :EndIf
              r←''
              :For i :In ⍳⍴,keys
                  THIS←i⊃,KEYS
                  this←i⊃,keys
                  :If 0∊⍴2⊃this
                      r,←⊂{({⍵⌿⍨0<↑∘⍴¨⍵[;1]}⍵._Data)[;2]}⍎'_',1⊃THIS
                  :Else
                      noOf←{1⊃⍴⍵._Data}⍎'_',1⊃THIS
                      :If noOf<where←(2⊃THIS){(Uppercase ⍵._Data[;1])⍳⊂Uppercase ⍺}⍎'_',1⊃THIS
                          :If 0=⎕NC'_default'
                              6 ⎕SIGNAL⍨'Unknown: ',2⊃this
                          :Else
                              r,←⊂_default
                          :EndIf
                      :Else
                          r,←where{⍵._Data[where;2]}⍎'_',1⊃THIS
                      :EndIf
                  :EndIf
              :EndFor
              :If ⍬≡⍴keys
                  :If 1=≡r
                      r←↑r
                  :Else
                      r←r[1]
                  :EndIf
              :EndIf
          :EndIf
        ∇
    :EndProperty

⍝⍝⍝ Constructors

    ∇ make0;parent
      ⍝ No parameter: Creates a new INI file by definition
      :Implements Constructor
      :Access Public
      Init ⍬
      Create''
    ∇

    ∇ make1(IniFilename)
      ⍝ Takes the name of one or more INI files to be processed as parameter.
      :Implements Constructor
      :Access Public
      Init ⍬
      :If 0 1∊⍨≡IniFilename
      :AndIf ~0∊⍴IniFilename
          IniFilename←,⊂IniFilename
      :EndIf
      :Trap 911
          Create IniFilename
      :Else
          ({(911≡⍵)∨⍬≡⍵:11 ⋄ ⍵}_EN)⎕SIGNAL⍨1⊃⎕DM
      :EndTrap
    ∇

    ∇ make2(IniFilename dummy)
      :Implements Constructor
      :Access Public
      ⍝ Takes the name of the INI file and what originally was supposed _
      ⍝ to be a reference to the "UnicodeFile" class as parameters. _
      ⍝ However, that property is not needed any longer (since version 1.6), _
      ⍝ and for that reason the argument "dummy" is simply ignored.
      Init ⍬
      Create IniFilename
    ∇

    ∇ Init dummy
      ⎕ML←3 ⋄ ⎕IO←1
      _import←0
      _EN←⍬
    ∇

    ∇ Create IniFilename;data;buffer2;bool;thisSection;this;ref;buffer;name;value;toc;val2;where;value2;sec;i;allRemarks;sectionRemarks;remark;TOC;NAME;si;import;⎕TRAP;en;dm;invalid
     
     ⍝ Called by the official constructors
      _changed←_oldStyleFlag←0
      _Sections←''
      ⎕TRAP←(0 'C' '_EN←⎕en ⋄ (⎕IO⊃⎕DM)⎕signal 911')
      :If ~0∊⍴_IniFilename←IniFilename
          _IniFilename←CheckExtension¨_IniFilename
          data←GetDataFrom _IniFilename
          data←{(0=+/b←(⎕UCS 9)=w←⍵):w ⋄ (b/w)←' ' ⋄ w}¨data ⍝ Transform TAB into blanks
          data←{0 1∊⍨≡⍵:⊂⍵ ⋄ ⍵}data
          data←{(0<∊⍴∘,¨⍵)⌿⍵}data
          buffer2←⊃data
          bool←{a∨≠\a←⍵=''''}buffer2
          ((,bool)/,buffer2)←' '
          allRemarks←(';'⍳⍨¨⊂[2]buffer2)↓¨data
          data←(¯1+';'⍳⍨¨⊂[2]buffer2)↑¨data
          data←dlb dtb data
          import←''
          (_oldStyleFlag dm en)←CheckValues data
          dm ⎕SIGNAL en
          :If 0<+/∨/¨'!Import'∘⍷¨data
              :Trap 911
                  (import data allRemarks)←Import data allRemarks
              :Else
                  911 ⎕SIGNAL⍨1⊃⎕DM
              :EndTrap
          :EndIf
          :Trap 911
              (data locals allRemarks)←GetLocals data allRemarks
          :Else
              911 ⎕SIGNAL⍨1⊃⎕DM
          :EndTrap
          :If ~0∊⍴import
              locals⍪←import,⊂''
          :EndIf
          :If ~0∊⍴data
              (data allRemarks)←(⊂+\'['=↑¨data)⊂¨data allRemarks
              :If ∨/invalid←(' '∊¨~∘'[]'∘↑¨data)∨¯1=↑∘⎕NC¨~∘'[ ]'∘↑¨data
                  11 ⎕SIGNAL⍨'Invalid section name(s): ',↑{⍺,', ',⍵}/invalid/↑¨data
              :EndIf
          :AndIf ~0∊⍴_Sections←'_',¨∪~∘'[ ]'∘↑¨data
              ⎕NS∘''¨sec←Uppercase _Sections
              (⍎¨sec)._Data←⊂0 4⍴''
              :For thisSection sectionRemarks :In ↓⍉⊃data allRemarks
                  this←'_','[ ]'~⍨1⊃thisSection
                  ref←⍎Uppercase this
                  ref.generalRemarks←⊂{⍵↓⍨+/∧\' '=⍵}1⊃sectionRemarks
                  buffer←{i←⍵⍳'=' ⋄ ((i-1)↑⍵)(i↓⍵)}¨1↓thisSection
                  :If OldStyleFlag
                  :AndIf 0∨.≠↑∘⎕NC¨↑¨buffer
                      ⎕TRAP←(0 1000)'S'
                      .
                  :EndIf
                  :If ∨/' '∊¨↑¨buffer
                      11 ⎕SIGNAL⍨'Invalid value name(s): ',↑{⍺,', ',⍵}/{⍵/⍨' '∊¨⍵}↑¨buffer
                  :EndIf
                  TOC←Uppercase toc←' '~¨⍨1⊃¨buffer
                  buffer←2⊃¨{2↑⍵,'' ''}¨buffer
                  EstablishValuesInSection(toc buffer)
                  :If 0∊⍴ref._Data
                      ref.Remarks←''
                  :Else
                      ref.Remarks←sectionRemarks~⊃,/{2=≡⍵:⍵ ⋄ ⊂⍵}¨ref._Data[;4]
                  :EndIf
              :EndFor
          :EndIf
      :EndIf
      ⎕DF(⍕⎕THIS),' on <',({0∊⍴⍵:⍵ ⋄ ⊃{⍺,';',⍵}/⍵}_IniFilename),'>'
    ∇

    ∇ EstablishValuesInSection(toc buffer);i;NAME;name;value;remark;value2;where
      :For i :In ⍳⍴,toc
          NAME←Uppercase name←i⊃toc
          value←i⊃buffer
          remark←{⍵↓⍨+/∧\' '=⍵}(1+i)⊃sectionRemarks ⍝ first one is the section
          :If 0∊⍴name
              ref._Data⍪←⍬ ⍬ ⍬ remark
          :Else
              :If 0∊⍴value
                  911 ⎕SIGNAL⍨'Invalid line: ',i⊃toc
              :ElseIf ∧/'{}'∊value
                  value2←⍎ReplaceCurlies(ref._Data[;1])value
              :Else
                  :If _oldStyleFlag
                      value2←value
                      value←'''',value,''''
                  :Else
                      value2←⍎value
                  :EndIf
                  :If ~0∊⍴value2
                      ('Invalid definition: ',name)⎕SIGNAL 911/⍨~0 1∊⍨≡value2
                  :EndIf
              :EndIf
              where←(Uppercase ref._Data[;1])⍳⊂NAME~','
              :If where>1⊃⍴ref._Data
                  ref._Data⍪←name value2 value remark
              :ElseIf ','∊name
                  :If 0∊⍴2⊃ref._Data[where;] ⍝ Is it the first time?
                      ref._Data[where;4]←⊂¨ref._Data[where;4]
                      ref._Data[where;3]←⊂''
                  :EndIf
                  ref._Data[where;2]←⊂(2⊃ref._Data[where;]),⊂value2
                  ref._Data[where;3]←⊂(3⊃ref._Data[where;]),⊂value
                  ref._Data[where;4]←⊂(4⊃ref._Data[where;]),⊂remark
              :Else
                  ref._Data[where;2 3 4]←value2 value remark
              :EndIf
          :EndIf
      :EndFor
    ∇

    ∇ r←{type}Convert r;data;thisSection;s;noOf;n;v;rf;allValues;this;theseNames;theseValues;code
      :Access Public
    ⍝ Takes a ref to a (typically empty) namespace and populates it with the values _
    ⍝ defined by the INI file entries.
    ⍝ If the optional left argument is "flat", sections are ignored and every entry _
    ⍝ gets a simple variable.
    ⍝ If it's not flat then section names are used as names for sub-namespaces.
    ⍝ Note that ""Convert"" will fail if the names used for sections and values _
    ⍝ are not proper APL names.
    ⍝ "Convert" injects a method ""List"" into the resulting namespace which _
    ⍝ prints a matrix to the session with all names and values.
      type←{0<⎕NC ⍵:⍎⍵ ⋄ ''}'type'
      'Left argument must be either empty or "flat"'⎕SIGNAL 11/⍨~(0∊⍴type)∨'flat'≡type
      data←Get ⍬ ⍬
      :If 'flat'≡type
          :If ~0∊⍴data←(0<↑∘⍴¨data[;2])⌿data                ⍝ Isolate pure data
              (theseNames theseValues)←↓⍉data[;2 3]         ⍝ All the names and values
              theseNames r.{⍎⍺,'←⍵'}¨theseValues            ⍝ Assign the values to the names
          :EndIf
          r.⎕FX'r←List;⎕IO;⎕ML' '(⎕IO ⎕ML)←0 3' 'r←{⍵,[.5]⍎¨⍵}⎕nl -2'  ⍝ Establish method "List"
      :Else
          allValues←data[;3]
          :For thisSection :In data[;1]~⊂''                 ⍝ Loop over sections
              s←data[;1]⍳⊂thisSection                       ⍝ Where does the section start?
              :If 0<noOf←{+/∧\0=↑¨⍴¨⍵}s↓data[;1]            ⍝ How many entries has the section?
                  rf←⍎'r.',thisSection,'←⎕NS'''''           ⍝ Create a namespace name after the section
                  rf.⎕DF thisSection                        ⍝ Set the display format
                  theseNames←noOf↑s↓data[;2]                ⍝ All the names and ...
                  theseValues←noOf↑s↓allValues              ⍝ ... values of this section
                  :If ¯1∊↑∘⎕NC¨theseNames
                      11 ⎕SIGNAL⍨'Invalid value name(s): ',{↑{⍺,', ',⍵}/⍵/⍨¯1=↑∘⎕NC¨⍵}theseNames
                  :EndIf
                  theseNames rf.{⍎⍺,'←⍵'}¨theseValues       ⍝ Assign values to names inside that namespace
              :EndIf
          :EndFor
          code←''
          code,←⊂' r←List section;toc;section_U;ind'
          code,←⊂'⍝ If "section" is empty or "*", all sections are returned'
          code,←⊂' ⎕IO←1 ⋄ ⎕ML←3'
          code,←⊂' r←'''''
          code,←⊂' :If (0∊⍴section)∨(,''*'')≡,section'
          code,←⊂'     r←⊃⍪/{⍵,⍵.{{⎕IO←0 ⋄ ⎕ML←3 ⋄ ⍵,[0.5]⍎¨⍵}⎕NL-2}⍵}¨⍎¨⎕NL-9'
          code,←⊂' :Else'
          code,←⊂'     ''Invalid right argument''⎕SIGNAL 11/⍨~(≡section)∊0 1'
          code,←⊂'     toc←⎕NL-9'
          code,←⊂'     section_U←#.APLTreeUtils.Uppercase section'
          code,←⊂'     :If (⊂section_U)∊toc'
          code,←⊂'         ind←toc⍳⊂section_U'
          code,←⊂'         r←{⍵.{{⎕IO←0 ⋄ ⎕ML←3 ⋄ ⍵,[0.5]⍎¨⍵}⎕NL-2}⍵}⍎ind⊃toc'
          code,←⊂'     :EndIf'
          code,←⊂' :EndIf'
          r.⎕FX code                                        ⍝ Establish method "List"
      :EndIf
    ∇

    ∇ r←{default}Get name;row;this;ref;subRow;where;thisSection;bool;section;key
    ⍝ Returns the (enclosed) value for a single value or a vector of values _
    ⍝ if more than one key was provided.
    ⍝ "name" might be either
    ⍝ <pre>('sectionName' 'key')</pre>
    ⍝ or
    ⍝ <pre>(('sectionName' 'key1')('sectionName' 'key2'))</pre>
    ⍝ or
    ⍝ <pre>'sectionName:key'</pre>
    ⍝ or
    ⍝ <pre>('sectionName:key1' 'sectionName:key2')</pre>
    ⍝ Note that mixed syntax is not supported:
    ⍝ <pre>('sectionName' ('key1' 'key2')) ⍝  invalid!
    ⍝ ('sectionName' 'key1') ''sectionName:key2' ⍝  invalid!</pre>
    ⍝ If "key" is empty, <b>all</b> values of that sections are returned.
    ⍝ If a requested value is not available, "default" (the left argument) _
    ⍝ is returned if specified, otherwise the property "Default" is returned _
    ⍝ if specified; otherwise an interrupt is signalled.
      :Access Public
      :If 0 1∊⍨≡name
          (section key)←{∨/b←':'=⍵:⍵{((¯1+⍵)↑⍺)(⍵↓⍺)}b⍳1 ⋄ ⍵}name
      :Else
          (section key)←name
      :EndIf
      'Invalid syntax!'⎕SIGNAL 11/⍨∧/2=¨≡¨section key
      section←{0∊⍴⍵:⍵ ⋄ '_',⍵}section
      :If ((⊂section)∊''⍬)∧(⊂key)∊''⍬
          r←0 3⍴''
          :For thisSection :In _SECTIONS
              r⍪←(⊂1↓thisSection),'' ''
              r⍪←(⊂''),thisSection⍎'_Data[;1 2]'
          :EndFor
      :ElseIf (⊂key)∊''⍬
          :If (1⊃⍴_Sections)≥where←(Uppercase _Sections)⍳Uppercase⊂section
              r←(where⊃_SECTIONS)⍎'_Data[;1 2]'
          :Else
              6 ⎕SIGNAL⍨'Unknown: "',(⍕∊section),'"'
          :EndIf
      :Else
          :If 0∊⍴section
              r←↑⍪/{⍵{(⊂⍺),⍵._Data}⍎⍵}¨_Sections
          :Else
              row←(Uppercase _Sections)⍳⊂Uppercase section
              :If (1⊃⍴_Sections)<row
                  :If (1⊃⍴_Sections)≥where←(Uppercase _Sections)⍳Uppercase⊂section
                      r←(Uppercase where⊃_Sections)⍎'_Data[;1 2]'
                  :Else
                      6 ⎕SIGNAL⍨'Unknown: "',(⍕∊1↓section),'"'
                  :EndIf
              :EndIf
              ref←⍎Uppercase this←row⊃_Sections
              :If 0∊⍴key
                  r←ref._Data
              :Else
                  key←{0 1∊⍨≡⍵:⊂⍵ ⋄ ⍵}key
                  subRow←(Uppercase ref._Data[;1])⍳Uppercase key
                  :If ∨/bool←(↑⍴ref._Data)<subRow
                      :If 0=⎕NC'default'
                          :If 0=⎕NC'Default'
                              6 ⎕SIGNAL⍨'Value Error: "',(⊃{⍺,',',⍵}/bool/key),'"'
                          :Else
                              :Trap 6
                                  r←(ref._Data[;2],⊂Default)[subRow]
                              :Else
                                  (1⊃⎕DM)⎕SIGNAL ⎕EN
                              :EndTrap
                          :EndIf
                      :Else
                          r←(ref._Data[;2],⊂default)[subRow]
                      :EndIf
                  :Else
                      r←ref._Data[subRow;2]
                  :EndIf
              :EndIf
          :EndIf
      :EndIf
      :If 2=⍴⍴r
          :Select 2⊃⍴r
          :Case 2
              r⌿⍨←0<↑∘⍴¨r[;1]
          :Case 3
              r⌿⍨←0∨.≠⍨↑∘⍴¨r
          :EndSelect
      :EndIf
    ∇

    ∇ data Put name;key;SECTION;where;ref;KEYS;buffer;KEY;toc;section;key
    ⍝ Set "name" to "data".
    ⍝ Note that "name" can be:
    ⍝ * A simple string. Must provide both, a "section" and a "name", _
    ⍝ separated by a colon:  "section:key"
    ⍝ * A nested vector of length 2.
    ⍝ Examples for the latter:
    ⍝ <pre>1⊃name ←→ section name </pre> _
    ⍝ <pre> 2⊃name ←→ key (or entry name)</pre>
      :Access Public
      :If 0 1∊⍨≡name
          'Must be specified as "sectionName:key"'⎕SIGNAL 11/⍨~':'∊name
          (section key)←{a←⍵⍳':' ⋄ (⍵↑⍨a-1)(a↓⍵)}name
      :Else
          (section key)←name
      :EndIf
      section←'_',section
      (SECTION KEY)←Uppercase(section key)
      :If (⊂SECTION)∊_SECTIONS
          where←_SECTIONS⍳⊂SECTION
      :Else
          _Sections,←⊂section
          where←⍴_Sections
      :EndIf
      :If 9=⎕NC where⊃_SECTIONS
          ref←⍎SECTION
      :Else
          6 ⎕SIGNAL⍨'Section unkown: ',SECTION,' (consider executing "AddSection")'
      :EndIf
      KEYS←Uppercase ref._Data[;1]
      :If (⎕DR data)∊11 83 645
          buffer←data(⍕data)
      :Else
          :If 2=≡data
              buffer←({ReplaceCurlies KEYS ⍵}¨data)({'''',⍵,''''}¨,data)
          :Else
              buffer←(ReplaceCurlies KEYS data)('''',data,'''')
          :EndIf
      :EndIf
      :If (⊂Uppercase key)∊KEYS
          where←KEYS⍳⊂Uppercase key
          :If ~_changed
              _changed←data≢2⊃ref._Data[where;]
          :EndIf
          ref._Data[where;2 3]←buffer
      :Else
          ref._Data←ref._Data,[1](⊂key),buffer,⊂'' ⍝ Avoid ,← syntax here - bug in Dyalog 2009-07
      :EndIf
    ∇

    ∇ {r}←Delete name;section;key;ref;toc;bool;this
      :Access Public
    ⍝ Delete a key or a section. Returns 1 if there was something to delete.
    ⍝ To delete "myKey" from "MySection":
    ⍝ <pre>('MySection' 'MyKey')</pre>
    ⍝ "name" can also be a nested vector:
    ⍝ <pre>(('SectionA' 'key1') ('SectionA' 'key2'))</pre>
    ⍝ or a full name like:
    ⍝ <pre>('section:key') or (('section:key1') ('section:key1'))</pre>
    ⍝ Note that ('section' ('key1' 'key2')) is <b>not</b> a valid syntax.
    ⍝ To delete an entire section:
    ⍝ <pre>Delete 'foo:'</pre>
    ⍝ If "name" is a vector of names, ""Delete"" calls itself recursively.
    ⍝ Mixed syntax is <b>not</b> supported, so use one of:
    ⍝ <pre>Delete ('sectiona' 'key1')('otherSection' 'key2')('thirdSection' '')
    ⍝ Delete ('sectiona:key1' 'otherSection:key2' 'thirdSection:')</pre>
      r←0
      :If 0 1∊⍨≡name
          'Invalid right argument'⎕SIGNAL 11/⍨1≠':'+.=name
          (section key)←' '~¨⍨2↑Uppercase{⍵⊂⍨':'≠⍵}name
          section←'_',section
          :If 0∊⍴key
              :If 9=⎕NC section
                  ⎕EX section
                  _Sections←(_SECTIONS≢¨⊂section)/_Sections
                  r←_changed←1
              :EndIf
          :Else
              key←{⍵∧.=' ':'' ⋄ ⍵}key
              (section key)←Uppercase{⍵⊂⍨':'≠⍵}name
              section←'_',section
              :If (⊂section)∊_SECTIONS
              :AndIf r←0<+/bool←(⊂key)≡¨toc←Uppercase{⍵._Data[;1]}ref←⍎section
                  ref._Data⌿⍨←~bool
                  r←_changed←1
              :EndIf
          :EndIf
      :Else
          :If 0 1∊⍨≡2⊃name
              :If 2=≡name
              :AndIf ~':'∊1⊃,name
                  r←Delete⊃{⍺,':',⍵}/name
              :Else
                  r←Delete¨name
              :EndIf
          :Else
              :If 3=≡name
              :AndIf 1 2≡≡¨name
                  r←Delete¨(⊂name[1]),¨⊂¨2⊃name
              :Else
                  r←⍬
                  :For this :In name
                      r,←Delete this
                  :EndFor
              :EndIf
          :EndIf
      :EndIf
    ∇

    ∇ r←Exist name;section;key;ref;this
      :Access Public
    ⍝ Returns 1 if "name" exists. "name" can be either a nested vector like _
    ⍝ <pre>('Section' 'key')</pre>
    ⍝ or
    ⍝ <pre>(('Section' 'key1') ('Section' 'key2'))</pre>
    ⍝ or a full name like
    ⍝ <pre>('section:key')</pre>
    ⍝  or
    ⍝ <pre>(('section:key1') ('section:key1'))</pre>
    ⍝ Note that ('section' ('key1' 'key2')) is NOT a valid syntax.
    ⍝ To check the existence of a section use
    ⍝ <pre>Exist 'foo:'</pre>
    ⍝ If "name" is a vector of names, ""Exist"" calls itself recursively.
    ⍝ Mixed syntax is NOT supported, so use one of:
    ⍝ <pre>Exist ('section' 'key1')('otherSection' 'key2')('thirdSection' '')
    ⍝ Exist ('section:key1' 'otherSection:key2' 'thirdSection:')</pre>
      :If 0 1∊⍨≡name
          (section key)←' '~¨⍨Uppercase 2↑{⍵⊂⍨':'≠⍵}name
          section←'_',section
          :If 0∊⍴key
              r←↑9=⎕NC section
          :Else
              key←{⍵∧.=' ':'' ⋄ ⍵}key
              :If 0∊⍴key
                  r←(⊂section)∊_SECTIONS
              :Else
                  :If r←(⊂section)∊_SECTIONS
                      ref←⍎section
                      r←(⊂key)∊Uppercase ref._Data[;1]
                  :EndIf
              :EndIf
          :EndIf
      :Else
          :If ∧/':'∊¨name
              r←Exist¨name
          :Else
              :If 3=≡name
                  'Mixed syntax not eligible'⎕SIGNAL 11/⍨(∨/':'∊¨name)∧(∨/2=≡¨name)
                  r←Exist¨name
              :Else
                  r←Exist⊃{⍺,':',⍵}/name
              :EndIf
          :EndIf
      :EndIf
    ∇

    ∇ {r}←AddSection name;NAME;ref
 ⍝ Adds a new section to the INI file.
 ⍝ Return a 1 if a section was added. A 0 is rerturned in case the section already exists.
      :Access Public
      NAME←Uppercase name←'_',name
      'Section name must be a valid APL name'⎕SIGNAL 11/⍨¯1=⎕NC NAME
      :If r←0=⎕NC NAME
          NAME ⎕NS''
          ref←⍎NAME
          ref._Data←0 4⍴''
          _Sections,←⊂NAME
          _changed←1
      :EndIf
    ∇

    ∇ {oldFilename}←Save filename;thisSection;thisKey;thisData;ref;buffer;origValue;THISSECTION;thisRemark;data;bool;where;i;this
      :Access Public
    ⍝ If "filename" is empty the INI is saved into the same file it originally came from.
    ⍝ Note that ""Save"" will throw a SYNTAX ERROR when...
    ⍝ * the instance was defined by more than one INI file.
    ⍝ * another INI file was imported.
      'An INI file with !Import statement cannot be saved'⎕SIGNAL 11/⍨_import
      :If 1<⍴_IniFilename
          11 ⎕SIGNAL⍨'This INI instance was defined by more than one INI file; it therefore cannot be saved.'
      :EndIf
      oldFilename←_IniFilename
      :If ~0∊⍴filename
          _IniFilename←,⊂filename
      :EndIf
      _IniFilename←CheckExtension¨_IniFilename
      :If 0=⎕NC'locals'
      :OrIf 0∊⍴locals
          data←⍬
      :Else
          bool←∧\0=↑∘⍴¨locals[;1]
          data←'; '∘,¨{⍵↓¨⍨+/∧\' '=⊃⍵}⊃¨bool⌿locals[;4]
          :For thisKey thisData origValue thisRemark :In ↓(~bool)⌿locals
              buffer←(thisKey{0∊⍴⍺:⍺ ⋄ ⍺,⍵}'='),{0=1↑0⍴⍵:⍕⍵ ⋄ '''',(DoubleQuotes ⍵),''''}thisData
              :If ~0∊⍴thisRemark
                  buffer,←thisKey{⍵,⍨(~0∊⍴⍺)/(⎕UCS 9)}'; ',thisRemark
              :EndIf
              data,←⊂buffer
          :EndFor
      :EndIf
      :For thisSection :In _Sections
          thisSection←1↓thisSection
          THISSECTION←Uppercase thisSection
          ref←⍎'_',THISSECTION
          :If 0=⎕NC'ref.generalRemarks'  ⍝ might not exist if the original filename was empty
              data,←⊂'[',THISSECTION,']'
          :Else
              data,←⊂'[',(THISSECTION),']',{⍵,⍨(~0∊⍴⍵)/(⎕UCS 9),'; '}1⊃1↑ref.generalRemarks
              data,←{⍵,⍨(~0∊⍴⍵)/'; '}¨1↓ref.generalRemarks
          :EndIf
          :For thisKey thisData origValue thisRemark :In ↓ref._Data
              buffer←''
              :If 0 1∊⍨≡thisData
                  :If '{'∊origValue
                      buffer←thisKey,'=''',(DoubleQuotes⍎origValue),''''
                  :Else
                      buffer←(thisKey{0∊⍴⍺:⍺ ⋄ ⍺,⍵}'='),{0=1↑0⍴⍵:⍕⍵ ⋄ '''',(DoubleQuotes ⍵),''''}thisData
                  :EndIf
                  :If ~0∊⍴thisRemark
                      buffer,←thisKey{⍵,⍨(~0∊⍴⍺)/(⎕UCS 9)}'; ',thisRemark
                  :EndIf
                  data,←⊂buffer
              :Else
             ⍝buffer←(⊂thisKey,'='''''),((⍴thisData)⍴⊂thisKey,',='),¨{0=1↑0⍴⍵:⍕⍵ ⋄ '''',(DoubleQuotes ⍵),''''}¨⍎¨origValue
                  buffer←(⊂thisKey,'='''''),(⊂thisKey,',='),¨{0=1↑0⍴⍵:⍕⍵ ⋄ '''',(DoubleQuotes ⍵),''''}¨⍎¨origValue
                  :If ~0∊⍴thisRemark
                      buffer,¨←{⍵,⍨(~0∊⍴⍵)/(⎕UCS 9),'; '}¨thisRemark
                  :EndIf
                  data,←buffer
              :EndIf
          :EndFor
      :EndFor
      bool←(';'≠↑¨1⍴¨data)∧(⎕UCS 9)∊¨data
      where←bool\{⍵⍳(⎕UCS 9)}¨bool/data
      buffer←(⎕UCS 9)⍴¨⍨{⍵-⍨⌈/1+⍵}⌊4÷⍨↑∘⍴¨(where-1)↑¨data
      :For i :In ⍳⍴,data
          :If i⊃bool
              this←i⊃data
              this[where[i]]←⊂i⊃buffer
              this←∊this
              (i⊃data)←this
          :EndIf
      :EndFor
      WriteUtf8File(1⊃_IniFilename)data
    ∇

    ∇ {r}←DeleteDefault
    ⍝ As soon as the "Default" property is set, one can get rid of it only be calling _
    ⍝ this method. Returns a 1 if there was a default, otherwise 0.
      :Access Public
      r←2=⎕NC'_default'
      ⎕EX'_default'
    ∇

    ∇ r←Help
      :Access Public shared
      r←''
      r,←⊂'To provide a list of methods and properties, use'
      r,←⊂'  #.ADOC.List ',⍕⎕THIS
      r,←⊂'To provide a full list of methods and properties, use'
      r,←⊂'  ''full'' #.ADOC.List ',⍕⎕THIS
      r,←⊂'To provide a fully-fledged documentation, use'
      r,←⊂'  #.ADOC.Browse ',⍕⎕THIS
      r←⊃r
    ∇

      fixkeys←{0 1∊⍨≡⍵:,⊂⍵
          ⍵}

    IsChar←{0 2∊⍨10|⎕DR ⍵}

    ∇ r←{extension}CheckExtension r
    ⍝ Add extension if there is no extension. Default is "ini"
      extension←{2=⎕NC ⍵:⍎⍵ ⋄ 'ini'}'extension'
      r←r,(~'.'∊{⍵↑⍨-⌊/'\/'⍳⍨⌽⍵}r)/'.',extension
    ∇

    ∇ r←_SECTIONS
      r←Uppercase _Sections
    ∇

    ∇ value←ReplaceCurlies(toc value);bool;val2;where;name;names;noOf
    ⍝ Replace "{foo}" with real data, if there are any curlies
      :If ∧/'{}'∊value
          bool←{{⍵∨≠\⍵}⍵∊'{}'}value
          names←Uppercase bool/value
          names←'}'~⍨¨'{'Split names
          names←(0<↑∘⍴¨names)/names
          :For name :In names
              :Trap ⍬
                  where←(Uppercase toc)⍳⊂name
                  :If (⍴toc)∨.<where
                      where←(Uppercase locals[;1])⍳⊂name
                      ('Invalid references found in INI file, check "',name,'"')⎕SIGNAL 6/⍨where∨.>1⊃⍴locals
                      val2←DoubleQuotes⊃locals[where;2]
                  :Else
                      val2←DoubleQuotes⊃ref._Data[where;2]
                  :EndIf
              :Else
                  val2←''
              :EndTrap
              noOf←¯1+value⍳'{'
              value←(-noOf)⌽val2,({⍵↓⍨⍵⍳'}'}noOf⌽value)
          :EndFor
          value←'''',(1↓¯1↓value),''''
      :EndIf
    ∇

      DoubleQuotes←{
          1<≡⍵:∇¨⍵
          ∨/b←''''=⍵:b{w←⍵ ⋄ (b/w)←⊂'''''' ⋄ ∊w}⍵
          ⍵
      }
    ∇ (data locals allRemarks)←GetLocals(data allRemarks);where;toc;i;name;value;remark;data2;allRemarks2;value2;itemNo;ref
⍝ Extract any "local" variabels.
⍝ Locals are defined as those assigned above the first section
      locals←0 4⍴⍬                                  ⍝ Initialize the main result
      :If '['∊∊data
          :If 0<where←¯1+'['⍳⍨↑¨data                    ⍝ Where starts the first section
              data2←where↑data                          ⍝ Take eveything above the first section
              data2←(+/∧\' '=⊃data2)↓¨data2             ⍝ Drop any leading blanks
              data2/⍨←'!Import'{((⍴⍺)↑[2]⊃⍵)∨.≠⍺}data2  ⍝ Remove all "!Import" directives
              data2←'='Split¨data2                      ⍝ Split remaining items at "=" (assignment)
              toc←' '~¨⍨1⊃¨data2                        ⍝ First part = names
              data2←2⊃¨{2↑⍵,'' ''}¨data2                ⍝ Second part = data
              allRemarks2←where↑allRemarks              ⍝
              :For i :In ⍳⍴,toc
                  name←i⊃toc
                  value←i⊃data2
                  remark←i⊃allRemarks2
                  :If 0∊⍴name
                      locals⍪←⍬ ⍬ ⍬ remark
                  :Else
                      :If ∧/'{}'∊value
                          value2←⍎ReplaceCurlies(toc value)
                      :Else
                          :If _oldStyleFlag
                              value2←value
                              value←'''',value,''''
                          :Else
                              value2←⍎value
                          :EndIf
                      :EndIf
                      itemNo←(Uppercase locals[;1])⍳⊂Uppercase name~','
                      :If itemNo>1⊃⍴locals
                          locals⍪←name value2 value remark
                      :ElseIf ','∊name
                          911 ⎕SIGNAL⍨'Invalid syntax: nested values are not supported for "local" variables'
                      :Else
                          locals[itemNo;2 3 4]←value2 value remark
                      :EndIf
                  :EndIf
              :EndFor
              locals[;4]←{2=≡⍵:⊃,/⍵ ⋄ ⍵}¨locals[;4]
              data←where↓data
              allRemarks←where↓allRemarks
          :EndIf
      :EndIf
    ∇

    ∇ r←name Process value;depth;buff;thisVal;i
    ⍝ Assign values to names
      depth←≡value
      :If depth∊0 1             ⍝ It's simple
          r←value
      :Else                     ⍝ It's nested!
          buff←''
          :For i :In ⍳⍴,value
              thisVal←i⊃value
              :If 0=1↑0⍴thisVal
                  buff,←',(⊂',(⍕thisVal),')'
              :Else
                  buff,←',(⊂''',(DoubleQuotes thisVal),''')'
              :EndIf
          :EndFor
          r←⍎buff
      :EndIf
    ∇

    ∇ r←GetDataFrom filenames
      r←⊃,/ReadUtf8File¨filenames
    ∇

      CheckValues←{
     ⍝ Returns either ('' ⍬) or ('Error msg' {number}).
     ⍝ Therefore the result can be used for a ⎕SIGNAL statement.
     ⍝ We don't have to worry about comments and trailing blanks: handled by now.
     ⍝ Side effect: sets global _oldStyleFlag
          data←⍵
          data←(0<↑∘⍴¨data)/data
          data←{⍵/⍨'['≠↑¨⍵}data             ⍝ We don't check section definitions here
          vals←{⍵↓⍨⍵⍳'='}¨data              ⍝ Remove the assigment as such.
          b1←''''=↑¨vals
          b2←''''=↑¨¯1⌽¨vals
          b3←{∧/1⊃⎕VFI ⍵}¨vals
          f←∨/(~b3)∧1=b1+b2
          dm←f{(0=⍺):'' ⋄ ⍺/'Invalid definition: ',1↓↑,/(⎕UCS 13),¨⍺/⍵}data
          en←f/11
          osf←∨/(~b3)∧{(~2∊⍵)∧0=⍵}b1+b2     ⍝ Old style flag
          (~0∊{(∧/1=⍵)∨(∧/0=⍵)}¨(~b3)∘/¨b1 b2):osf dm en
          dm←'IniFile is inconsistent'
          en←11
          osf dm en
      }

:EndClass ⍝ IniFiles