﻿:Namespace App2
⍝ Reworked "zipweather" MiServer Demo "Demo\zipweather.dyalog"
⍝ Ends of Plot functions reworked to just call ch.Close
⍝ Requires SAWS and ZipWeatherWs from "Demo\Code" to have been loaded into root

    (⎕ML ⎕IO)←0 1 ⍝ Avoid inheritance from environment

    ∇ OK x
    ⍝ General Purpose Return-code checker
    (⍕x) ⎕SIGNAL (0≠1⊃x)/11
    ∇ 

    ∇ r←Init;z
      #.⎕CY 'conga' ⍝ Conga into the root
      #.⎕CY 'sqapl' ⍝ SQAPL too
      ⎕CY 'rainpro' ⍝ Rainpro we bring into this NS
      OK #.DRC.Init''
      OK #.SQA.Init''
      OK #.SQA.Connect 'ZipCodes' '' ('DriverOptions' 'DRIVER={Microsoft Access Driver (*.mdb, *.accdb)};DBQ=%SiteRoot%\data\zipcodes.accdb;ExtendedAnsiSQL=1;MaxBufferSize=2048;')
      ⍝ ↑↑↑ From Miserver\Config\Datasources.xml
      OK (rc States)←2↑#.SQA.Do 'ZipCodes' 'select * from States order by StateName'
      OK #.SQA.Prepare 'ZipCodes.ZCQ' 'select * from ZipCodes where StateAbbr = :a<C2: order by Zipcode'
    ∇
    
    ∇ Start;z;data;⎕TRAP
      ⎕TRAP←DrA.TrapGUI
      'F1'⎕WC'Form' 'Dyalog 2012 Weather Forecaster' ('Size' 600 800)('Coord' 'Pixel')
      'F1.State' ⎕WC 'Combo' States (10 10)(⍬ 100)('Event' 'Select' 'SelectState')
      'F1.Zip' ⎕WC 'Combo' (0⍴⊂'') (10 120)(⍬ 100)('Event' 'Select' 'SelectZip')
      'F1.C1' ⎕WC 'SubForm' (50 10)(545 385)
      'F1.C1' ⎕WS 'Coord' 'User'('Yrange' 385)('Xrange' 545)
      'F1.C2' ⎕WC 'SubForm' (50 405)(545 385)
      'F1.C2' ⎕WS 'Coord' 'User'('Yrange' 385)('Xrange' 545)
      ch.New 385 545 ⍝ Initialize RainPro
      'F1.TXT' ⎕WC 'Text' ⍬ (30 10)
      ⎕DQ'F1'
      ⍝ ⎕OFF
    ∇

    ∇ SelectState msg
      ∘
    ∇

    ∇ SelectZip msg
      zip←∘
          :If 0≠⊃data←#.ZipWeatherWS.WeatherData zip
              F1.TXT←'Enclose'Could not retrieve weather for "',zip,'"<br/>',3⊃data
          :Else
              F1.TXT←'Weather for ',⍕zip
              Weather←Extract 2⊃data
              PostScrp.PS (Plot1 Weather) '#.App2.F1.C1' 1 PostScrp.∆cmap
              PostScrp.PS (Plot2 Weather) '#.App2.F1.C2' 1 PostScrp.∆cmap
          :EndIf
    ∇

    ∇ Weather←Extract data
      Weather←⎕NS''
      Weather.(Date Temp Dewpt AppTemp RHum SkyCover WindSpeed WindDir)←↓[1]data
      Weather.Date←cal_ts2ord¨Weather.Date
    ∇

    ∇ PG←Plot1 Weather;temps;tn;file
     ⍝ Simple timeseries of various temperature records
      ch.SetHeading'Temperature Series'
      ch.SetStyle'XYPlot'
     
      ch.SetMarginRight 36
     
      ch.SetPenWidths 1.2
      ch.SetLineStyles LineStyle.(Solid Dash Solid)
      ch.SetFillStyles 15
     
      ch.SetXAxisStyle'Date,Grid'
      ch.SetYCaption'Deg C'
      ch.SetYAxisStyle'AtEnd'
      temps←{⎕ML←1 ⋄ ∊⍵}Weather.(Temp AppTemp)
      ch.SetYRange(0⌊¯5+⌊/temps),5+⌈/temps
     
      ch.Plot Weather.(Date Temp AppTemp)
     
      ch.SetMarginTop 220  ⍝ Shorter sec axis!
      ch.Ysec
     
      ch.SetXAxisStyle'Date,nogrid'
      ch.SetYCaption'%'
      ch.SetYAxisStyle'Right,AtEnd,ForceZero,grid'
      ch.SetYRange 0 100
      ch.SetYTickMarks 50
     
      ch.SetBarWidths 0.06 ⍝ Days
      ch.SetBarEdgeNib 0
     
      ch.Xbar Weather.(Date RHum)
     
      ch.SetKeyText'Temperature,Apparent Temperature,Relative Humidity'
      PG←ch.Close
    ∇

    ∇ html←Plot2 Weather;file;tn
      ch.SetMargins 54 36 12 12 ⍝ tblr
     
      ch.SetHeading'Wind Directions and Speeds'
     
      ch.SetMarkers Marker.SmallBullet
     
      ch.SetStyle'XYplot,markers,risers,nolines,Compass'
      ch.SetYAxisStyle'Forcezero,Gridlines'
      ch.SetLabelStyle'opaque'
      ch.SetXTickMarks 45
      ch.Polar Weather.WindDir,[1.5]Weather.WindSpeed
     
      PG←ch.Close
    ∇

    ∇ R←cal_ts2ord TS;DT;M;∆YR;∆MO;ENT;TD;LY;VALID;⎕ML;⎕IO;limit;xl;wdw
      ⍝ Convert ⎕TS to ordinal days (decimal part is accurate to ~1sec)
      ⍝ Uses the Excel convention (based 1899 - 1900 IS a leap year)
      ⎕IO←⎕ML←1 ⋄ DT←3↑TS
      LY←(0 0 0 1),400⍴(196⍴1 0 0 0),0 0 0 0
      ∆YR←0,(+\404⍴365)++\LY
      ∆MO←0 31 59 90 120 151 181 212 243 273 304 334 365
      DT[1]←DT[1]-1900
      ⍝ RANGE CHECK ON YR/MTH/DAY
      VALID←DT∧.<400 13 32 ⋄ R←¯1 ⋄ →VALID↓0
      ⍝ Days in month check fixed .. ↓↓↓↓↓↓↓↓ ... Feb 99 ACDS
      VALID←VALID∧DT[3]≤(LY[1+DT[1]]×2=DT[2])+-/∆MO[13⌊1⌈DT[2]∘.+1 0] ⋄ →VALID↓0
      ⍝FINALLY ADD UP CUMULATIVE DAYS BASED ON 1900.
      R←∆YR[1⌈DT[1]+1]+∆MO[1⌈DT[2]]+DT[3]
      ⍝ AND ADD ON ONE DAY FOR LEAP-YEARS
      R←R+(DT[2]>2)∧LY[1+DT[1]]
      ⍝ Now the decimal part
      R←R+0.00001×⌈0.864÷⍨60 60 60⊥TS[4 5 6]
    ∇

:EndNamespace